"""
ffmpeg_streaming.media
~~~~~~~~~~~~

Size and Bitrate Objects


:copyright: (c) 2020 by Amin Yazdanpanah.
:website: https://www.aminyazdanpanah.com
:email: contact@aminyazdanpanah.com
:license: MIT, see LICENSE for more details.
"""
import math

OVERALL_TO_VIDEO_COEFFICIENT = 1
MAX_RATE_COEFFICIENT = 1.2
BUFFER_SIZE = 65536


def cnv_bitrate(bitrate: int, _type: str) -> str:
    if _type == "k":
        bitrate = round(bitrate / 1024)
    elif _type == "m":
        bitrate = round(bitrate / 1024 * 1024)
    else:
        raise ValueError("Unknown type!")

    return str(bitrate) + _type


class Bitrate:
    def __init__(self, video: int = None, audio: int = None, overall: int = None, **kwargs):
        """
        @TODO: add documentation
        """
        if video is None and overall is None:
            raise ValueError("You must at least specify value of the video or overall format")
        self.overall_ = overall
        self.video_ = video
        self.audio_ = audio
        self.kwargs = kwargs
        self.type = kwargs.pop("type", "k")

    @property
    def overall(self):
        """
        @TODO: add documentation
        """
        return cnv_bitrate(self.overall_, self.type) if self.overall_ is not None else None

    @property
    def video(self):
        """
        @TODO: add documentation
        """
        return cnv_bitrate(self.video_, self.type) if self.video_ is not None else None

    @property
    def audio(self):
        """
        @TODO: add documentation
        """
        return cnv_bitrate(self.audio_, self.type) if self.audio_ is not None else 'copy'

    def calc_video(self, convert: bool = True):
        """
        @TODO: add documentation
        """
        if self.video_ is not None and self.video_ != 0:
            val = self.video_
        else:
            val = int(self.overall_ * OVERALL_TO_VIDEO_COEFFICIENT)

        return cnv_bitrate(val, self.type) if convert else val

    @property
    def calc_overall(self):
        """
        @TODO: add documentation
        """
        return self.overall_ if self.overall_ is not None else self.video_ + self.audio_


def multiple_up(value, multiple):
    while 0 != value % multiple:
        value += 1

    return value


def multiple_down(value, multiple):
    while 0 != value % multiple:
        value -= 1

    return value


class Ratio:
    def __init__(self, width: int, height: int):
        """
        @TODO: add documentation
        """
        self.width = width
        self.height = height

    def get_value(self) -> float:
        return self.width / self.height

    def calculate_width(self, height: int, multiple: int = 1) -> int:
        """
        @TODO: add documentation
        """
        max_w = multiple_up(math.ceil(self.get_value() * height), multiple)
        min_w = multiple_down(math.floor(self.get_value() * height), multiple)

        max_r = abs(self.get_value() - (max_w / height))
        min_r = abs(self.get_value() - (min_w / height))

        return max_w if max_r < min_r else min_w

    def calculate_height(self, width: int, multiple: int = 1) -> int:
        """
        @TODO: add documentation
        """
        max_h = multiple_up(math.ceil(width / self.get_value()), multiple)
        min_h = multiple_down(math.floor(width / self.get_value()), multiple)

        max_r = abs(self.get_value() - (width / max_h))
        min_r = abs(self.get_value() - (width / min_h))

        return max_h if max_r < min_r else min_h


class Size:
    def __init__(self, width: int, height: int):
        """
        @TODO: add documentation
        """
        self.width = width
        self.height = height

    @property
    def ratio(self) -> Ratio:
        """
        @TODO: add documentation
        """
        return Ratio(self.width, self.height)

    def __str__(self) -> str:
        """
        @TODO: add documentation
        """
        return f"{self.width}x{self.height}"


__all__ = [
    'Size',
    'Bitrate'
]


import json
import os
import unittest

import ffmpeg_streaming
from ffmpeg_streaming import Formats
from ffmpeg_streaming import FFProbe, Representation, Size, Bitrate
from ffmpeg_streaming._media import HLS, DASH
from ffmpeg_streaming.ffprobe import Streams


class TestStreaming(unittest.TestCase):

    def setUp(self):
        self.src_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'files')
        self.src_video = os.path.join(self.src_dir, 'test.mp4')
        with open(os.path.join(self.src_dir, 'fixture_ffprobe')) as ffprobe:
            self.ffprobe = json.loads(ffprobe.read())

    def test_ffprobe(self):
        ffprobe = FFProbe(self.src_video)

        all_media = ffprobe.all()

        streams = ffprobe.streams()
        self.assertIsInstance(streams, Streams)
        self.assertEqual(streams.all(), all_media['streams'])
        self.assertEqual(streams.first_stream(), all_media['streams'][0])
        self.assertEqual(streams.video()['codec_type'], 'video')
        self.assertEqual(streams.audio()['codec_type'], 'audio')

    def test_hls(self):
        video = ffmpeg_streaming.input(self.src_video)
        hls_obj = video.hls(Formats.h264())
        self.assertIsInstance(hls_obj, HLS)
        self.assertEqual(hls_obj.media.input, self.src_video)

        hls_obj.representations(Representation(Size(256, 144), Bitrate(102400)))
        rep_1 = hls_obj.reps[0]
        self.assertIsInstance(rep_1, Representation)
        self.assertEqual(str(rep_1.size), '256x144')
        self.assertEqual(rep_1.bitrate.video, '100k')

        hls_obj.auto_generate_representations()
        reps = list(hls_obj.reps)
        self.assertEqual(len(reps), 3)

        for rep_ in reps:
            self.assertIsInstance(rep_, Representation)
        self.assertEqual(str(reps[0].size), '480x270')
        self.assertEqual(reps[0].bitrate.video, '176k')
        self.assertEqual(str(reps[1].size), '426x240')
        self.assertEqual(reps[1].bitrate.video, '88k')
        self.assertEqual(str(reps[2].size), '256x144')
        self.assertEqual(reps[2].bitrate.video, '71k')

        hls_obj.output(os.path.join(self.src_dir, 'hls', 'test.m3u8'))
        with open(os.path.join(self.src_dir, 'fixture_test.m3u8')) as test_m3u8:
            expected_m3u8 = test_m3u8.read()
        with open(os.path.join(self.src_dir, 'hls', 'test.m3u8')) as test_m3u8:
            actual_m3u8 = test_m3u8.read()
        self.assertEqual(actual_m3u8, expected_m3u8)
        with open(os.path.join(self.src_dir, 'hls', 'test_270p.m3u8')) as test_m3u8:
            actual_270_m3u8 = test_m3u8.readlines()
        self.assertEqual(actual_270_m3u8[0].replace('\n', ''), '#EXTM3U')
        self.assertEqual(actual_270_m3u8[1].replace('\n', ''), '#EXT-X-VERSION:3')
        self.assertEqual(actual_270_m3u8[2].replace('\n', ''), '#EXT-X-ALLOW-CACHE:YES')

    def test_encrypted_hls(self):
        video = ffmpeg_streaming.input(self.src_video)
        hls_obj = video.hls(Formats.h264())
        hls_obj.encryption(os.path.join(self.src_dir, 'enc.key'), 'https://www.aminyazdanpanah.com/enc.key')

        self.assertIsNotNone(hls_obj.options.get('hls_key_info_file', None))

        with open(hls_obj.options.get('hls_key_info_file', None)) as key_info:
            key_info = key_info.readlines()
        self.assertEqual(key_info[0].replace('\n', ''), 'https://www.aminyazdanpanah.com/enc.key')
        self.assertEqual(key_info[1].replace('\n', ''), os.path.join(self.src_dir, 'enc.key'))

        hls_obj.auto_generate_representations()

        hls_obj.output(os.path.join(self.src_dir, 'encrypted_hls', 'test.m3u8'), stderr=False)
        with open(os.path.join(self.src_dir, 'fixture_test.m3u8')) as test_m3u8:
            expected_m3u8 = test_m3u8.read()
        with open(os.path.join(self.src_dir, 'encrypted_hls', 'test.m3u8')) as test_m3u8:
            actual_encrypted_m3u8 = test_m3u8.read()
        self.assertEqual(actual_encrypted_m3u8, expected_m3u8)

    def test_dash(self):
        video = ffmpeg_streaming.input(self.src_video)
        dash_obj = video.dash(Formats.hevc())
        self.assertIsInstance(dash_obj, DASH)
        self.assertEqual(dash_obj.media.input, self.src_video)

        dash_obj.auto_generate_representations()

        dash_obj.output(os.path.join(self.src_dir, 'dash', 'test.mpd'), stderr=False)
        with open(os.path.join(self.src_dir, 'dash', 'test.mpd')) as test_mpd:
            actual_mpd = test_mpd.readlines()
        self.assertEqual(actual_mpd[0].replace('\n', ''), '<?xml version="1.0" encoding="utf-8"?>')
        self.assertEqual(actual_mpd[1].replace('\n', ''), '<MPD xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"')


if __name__ == '__main__':
    unittest.main()



from ._input import *
from ._format import *
from ._reperesentation import *
from ._media_property import *
from ._clouds import *
from .ffprobe import *

__all__ = _input.__all__ + _format.__all__ + _reperesentation.__all__ + _media_property.__all__ + _clouds.__all__ + ffprobe.__all__



"""
ffmpeg_streaming.media
~~~~~~~~~~~~

Media object to build a stream objects


:copyright: (c) 2020 by Amin Yazdanpanah.
:website: https://www.aminyazdanpanah.com
:email: contact@aminyazdanpanah.com
:license: MIT, see LICENSE for more details.
"""
import abc
import os
import shutil
import tempfile
import atexit
import asyncio
import copy

from ._clouds import CloudManager
from ._command_builder import command_builder
from ._format import Format
from ._hls_helper import HLSKeyInfoFile, HLSMasterPlaylist
from ._process import Process
from ._reperesentation import Representation, AutoRep
from ._utiles import mkdir, rm
from .ffprobe import FFProbe


class Save(abc.ABC):
    def __init__(self, media, _format: Format, **options):
        """
        @TODO: add documentation
        """
        atexit.register(self.finish_up)

        self.output_ = ''
        self.key = None
        self.media = media
        self.format = _format
        self.options = options
        self.pipe = None
        self.output_temp = False

    def finish_up(self):
        """
        @TODO: add documentation
        """
        if self.media.input_temp:
            rm(self.media.input)
        if self.output_temp:
            self.clouds.transfer('upload_directory', os.path.dirname(self.output_))
            if self.output:
                shutil.move(os.path.dirname(self.output_), os.path.dirname(str(self.output)))
            else:
                shutil.rmtree(os.path.dirname(str(self.output_)), ignore_errors=True)

    @abc.abstractmethod
    def set_up(self):
        pass

    def __getattr__(self, name):
        def method(*args, **kwargs):
            if name in ['save', 'package']:
                self.output(*args, **kwargs)
            else:
                raise AttributeError("The object has no attribute {}".format(name))

        return method

    def output(self, output: str = None, clouds: CloudManager = None,
               run_command: bool = True, ffmpeg_bin: str = 'ffmpeg', monitor: callable = None, **options):
        """
        @TODO: add documentation
        """
        if output is None and clouds is None:
            self.output_ = self.media.input
        elif clouds is not None:
            self.output_temp = True
            setattr(self, 'clouds', clouds)
            setattr(self, 'output', output)
            if clouds.filename is None:
                clouds.filename = os.path.basename(output if output is not None else self.media.input)
            self.output_ = os.path.join(tempfile.mkdtemp(prefix='ffmpeg_streaming_'), clouds.filename)
        else:
            mkdir(os.path.dirname(output))
            self.output_ = output

        self.set_up()

        if run_command:
            self.run(ffmpeg_bin, monitor, **options)

    def probe(self, ffprobe_bin='ffprobe') -> FFProbe:
        """
        @TODO: add documentation
        """
        return FFProbe(self.media.input, ffprobe_bin)

    def _run(self, ffmpeg_bin, monitor: callable = None, **options):
        """
        @TODO: add documentation
        """
        with Process(self, command_builder(ffmpeg_bin, self), monitor, **options) as process:
            self.pipe, err = process.run()

    async def async_run(self, ffmpeg_bin, monitor: callable = None, **options):
        """
        @TODO: add documentation
        """
        self._run(ffmpeg_bin, monitor, **options)

    def run(self, ffmpeg_bin, monitor: callable = None, **options):
        """
        @TODO: add documentation
        """
        async_run = options.pop('async_run', True)

        if async_run:
            asyncio.run(self.async_run(ffmpeg_bin, monitor, **options))
        else:
            self._run(ffmpeg_bin, monitor, **options)


class Streaming(Save, abc.ABC):
    def __init__(self, media, _format: Format, **options):
        """
        @TODO: add documentation
        """
        self.reps = list
        super(Streaming, self).__init__(media, _format, **options)

    def representations(self, *reps: Representation):
        self.reps = list(reps)

    def auto_generate_representations(self, heights=None, bitrate=None, ffprobe_bin='ffprobe', include_original=True,
                                      ascending_sort=False):
        """
        @TODO: add documentation
        """
        probe = self.probe(ffprobe_bin)
        self.reps = AutoRep(probe.video_size, probe.bitrate, self.format, heights, bitrate, include_original)

        if ascending_sort:
            self.reps = sorted(self.reps, key=lambda rep: rep.bitrate.overall_)

    def add_filter(self, *_filter: str):
        """
        @TODO: add documentation
        """
        _filters = self.options.pop('filter_complex', None)
        _filters = _filters + "," + ",".join(list(_filter)) if _filters is not None else ",".join(list(_filter))
        self.options.update({'filter_complex': _filters})

    def watermarking(self, path, _filter='overlay=10:10'):
        """
        @TODO: add documentation
        """
        self.media.inputs.input(path)
        self.add_filter(_filter)


class DASH(Streaming):
    def set_up(self):
        pass

    def generate_hls_playlist(self):
        self.options.update({'hls_playlist': 1})


class HLS(Streaming):
    KEY_INFO_FILE_PATH = None
    PERIODIC_RE_KEY_FLAG = 'periodic_rekey'
    MASTER_PLAYLIST_IS_SAVED = False

    def set_up(self):
        """
        @TODO: add documentation
        """
        if not HLS.MASTER_PLAYLIST_IS_SAVED:
            self.save_master_playlist()

    def encryption(self, path: str, url: str, key_rotation_period: int = 0, needle: str = ".ts' for writing", length: int = 16):
        """
        @TODO: add documentation
        """
        with tempfile.NamedTemporaryFile(mode='w', suffix='_py_ff_vi_st.tmp', delete=False) as key_info:
            HLS.KEY_INFO_FILE_PATH = key_info.name

        key_info_file = HLSKeyInfoFile(HLS.KEY_INFO_FILE_PATH, path, url, key_rotation_period, needle, length)
        self.options.update({'hls_key_info_file': str(key_info_file)})

        if key_rotation_period > 0:
            setattr(self, 'key_rotation', key_info_file)
            self.flags(HLS.PERIODIC_RE_KEY_FLAG)

    def fragmented_mp4(self):
        """
        @TODO: add documentation
        """
        self.options.update({'hls_segment_type': 'fmp4'})

    def save_master_playlist(self, path=None):
        """
        @TODO: add documentation
        """
        if path is not None:
            HLS.MASTER_PLAYLIST_IS_SAVED = True

        HLSMasterPlaylist.generate(self, path)

    def flags(self, *flags: str):
        """
        @TODO: add documentation
        """
        hls_flags = self.options.pop('hls_flags', None)
        hls_flags = hls_flags + "+" + "+".join(list(flags)) if hls_flags is not None else "+".join(list(flags))
        self.options.update({'hls_flags': hls_flags})

    def finish_up(self):
        """
        @TODO: add documentation
        """
        if HLS.KEY_INFO_FILE_PATH is not None:
            rm(HLS.KEY_INFO_FILE_PATH)

        super(HLS, self).finish_up()


class Stream2File(Save):
    """
    @TODO: add documentation
    """
    def set_up(self):
        pass


class Media(object):
    def __init__(self, _inputs):
        """
        @TODO: add documentation
        """
        self.inputs = _inputs

        first_input = dict(copy.copy(_inputs.inputs[0]))
        self.input = first_input.get('i', None)
        self.input_temp = first_input.get('is_tmp', False)

    def hls(self, _format: Format, **hls_options):
        """
        @TODO: add documentation
        """
        return HLS(self, _format, **hls_options)

    def dash(self, _format: Format, **dash_options):
        """
        @TODO: add documentation
        """
        return DASH(self, _format, **dash_options)

    def stream2file(self, _format: Format, **options):
        """
        @TODO: add documentation
        """
        return Stream2File(self, _format, **options)



"""
ffmpeg_streaming.input
~~~~~~~~~~~~

Input options


:copyright: (c) 2020 by Amin Yazdanpanah.
:website: https://www.aminyazdanpanah.com
:email: contact@aminyazdanpanah.com
:license: MIT, see LICENSE for more details.
"""
from ffmpeg_streaming._media import Media
from ffmpeg_streaming._utiles import get_os, cnv_options_to_args
from ffmpeg_streaming._clouds import Clouds

cloud = None


class Capture(object):
    def __init__(self, video, options):
        """
        @TODO: add documentation
        """
        self.options = options
        self.video = video

    def _linux(self):
        is_screen = self.options.pop('screen', False)
        if is_screen:
            cap = 'x11grab'
        else:
            cap = 'v4l2'

        return {
            'f': cap,
            'i': self.video
        }

    def _windows(self):
        self.video = 'video=' + str(self.video)
        windows_audio = self.options.pop('windows_audio', None)
        if windows_audio is not None:
            self.video = self.video + ':audio=' + str(windows_audio)

        return {
            'f': 'dshow',
            'i': self.video
        }

    def _os_x(self):
        return {
            'f': 'avfoundation',
            'i': self.video
        }

    @staticmethod
    def _unknown():
        raise OSError("Unreported OS!")

    def __iter__(self):
        yield from getattr(self, '_' + get_os())().items()


def get_from_cloud(_cloud: Clouds, options: dict):
    """
    @TODO: add documentation
    """
    global cloud
    if cloud is None:
        save_to = options.pop('save_to', None)
        cloud = {
            'i': _cloud.download(save_to, **options),
            'is_tmp': True if save_to is None else False
        }

    return cloud


class InputOption(object):
    def __init__(self, _input, **options):
        """
        @TODO: add documentation
        """
        self.input_ = _input
        self.options = options

    def __str__(self):
        """
        @TODO: add documentation
        """
        return " ".join(cnv_options_to_args(self._create()))

    def __iter__(self):
        """
        @TODO: add documentation
        """
        yield from self._create().items()

    def _create(self):
        """
        @TODO: add documentation
        """
        options = self.options.pop('pre_opts', {'y': None})
        is_cap = self.options.pop('capture', False)

        if isinstance(self.input_, Clouds):
            options.update(get_from_cloud(self.input_, self.options))
        elif is_cap:
            options.update(Capture(self.input_, self.options))
        elif isinstance(self.input_, (str, int)):
            i_options = {'i': str(self.input_)}
            i_options.update(self.options)
            options.update(i_options)
        else:
            raise ValueError("Unknown input!")

        return options


class Input:
    def __init__(self, _input: InputOption):
        """
        @TODO: add documentation
        """
        self.inputs = [_input]

    def input(self, _input, **options):
        """
        @TODO: add documentation
        """
        self.inputs.append(InputOption(_input, **options))

    def __getattr__(self, name):
        """
        @TODO: add documentation
        """
        def method(*args, **kwargs):
            media = Media(self)
            if hasattr(media, name):
                return getattr(media, name)(*args, **kwargs)
            else:
                raise AttributeError("The object has no attribute {}".format(name))

        return method


def input(_input, **options) -> Input:
    """Input options (ffmpeg pre_option ``-i`` input options)
        You can also pass a cloud object as an input to the method. the file will be downloaded and will pass it to ffmpeg
        if you want to open a resource from a pipe, set input "pipe:"
        if you want to open a resource from a capture device, pass a device name as filename and set the capture keyword
        to True. To list the supported, connected capture devices, see https://trac.ffmpeg.org/wiki/Capture/Webcam
         and https://trac.ffmpeg.org/wiki/Capture/Desktop. See https://ffmpeg.org/ffmpeg.html#Main-options and
         https://ffmpeg.org/ffmpeg-protocols.html for more information about input option and supported resources
         such as http, ftp, and so on.
        """
    return Input(InputOption(_input, **options))


__all__ = [
    'input',
]



"""
ffmpeg_streaming.clouds
~~~~~~~~~~~~

Upload and download files -> clouds


:copyright: (c) 2020 by Amin Yazdanpanah.
:website: https://www.aminyazdanpanah.com
:email: contact@aminyazdanpanah.com
:license: MIT, see LICENSE for more details.
"""

import abc
import logging
import tempfile
from os import listdir
from os.path import isfile, join, basename


class Clouds(abc.ABC):
    """
    @TODO: add documentation
    """
    @abc.abstractmethod
    def upload_directory(self, directory: str, **options) -> None:
        pass

    @abc.abstractmethod
    def download(self, filename: str = None, **options) -> str:
        pass


class S3(Clouds):
    def __init__(self, **options):
        """
        @TODO: add documentation
        """
        try:
            import boto3
            from botocore.exceptions import ClientError
        except ImportError as e:
            raise ImportError("No specified import name! make sure that you have installed the package via pip:\n\n"
                              "pip install boto3")

        self.s3 = boto3.client('s3', **options)
        self.err = ClientError

    def upload_directory(self, directory, **options):
        bucket_name = options.pop('bucket_name', None)
        folder = options.pop('folder', '')
        if bucket_name is None:
            raise ValueError('You should pass a bucket name')

        files = [f for f in listdir(directory) if isfile(join(directory, f))]

        try:
            for file in files:
                self.s3.upload_file(join(directory, file), bucket_name, join(folder, file).replace("\\", "/"))
        except self.err as e:
            logging.error(e)
            raise RuntimeError(e)

        logging.info("The {} directory was uploaded to Amazon S3 successfully".format(directory))

    def download(self, filename=None, **options):
        bucket_name = options.pop('bucket_name', None)
        key = options.pop('key', None)

        if bucket_name is None or key is None:
            raise ValueError('You should pass a bucket and key name')

        if filename is None:
            filename = tempfile.NamedTemporaryFile(prefix=basename(key), delete=False)
        else:
            filename = open(filename, 'wb')

        try:
            with filename as f:
                self.s3.download_fileobj(bucket_name, key, f)
            logging.info("The " + filename.name + " file was downloaded")
        except self.err as e:
            logging.error(e)
            raise RuntimeError(e)

        return filename.name


class GCS(Clouds):
    CLIENT = None

    def __init__(self, **options):
        """
        @TODO: add documentation
        """
        try:
            from google.cloud import storage
        except ImportError as e:
            raise ImportError("No specified import name! make sure that you have installed the package via pip:\n\n"
                              "pip install google-cloud-storage")
        GCS.CLIENT = storage.Client(**options)

    def upload_directory(self, directory, **options):
        bucket_name = options.pop('bucket_name', None)
        if bucket_name is None:
            raise ValueError('You should pass a bucket name')

        bucket = GCS.CLIENT.get_bucket(bucket_name)
        folder = options.pop('folder', '')
        files = [f for f in listdir(directory) if isfile(join(directory, f))]

        for file in files:
            blob = bucket.blob(join(folder, file).replace("\\", "/"), **options)
            blob.upload_from_filename(join(directory, file))

    def download(self, filename=None, **options):
        bucket_name = options.pop('bucket_name', None)
        if bucket_name is None:
            raise ValueError('You should pass a bucket name')

        bucket = GCS.CLIENT.get_bucket(bucket_name)
        object_name = options.pop('object_name', None)

        if object_name is None:
            raise ValueError('You should pass an object name')

        if filename is None:
            with tempfile.NamedTemporaryFile(prefix=basename(object_name), delete=False) as tmp:
                filename = tmp.name

        blob = bucket.get_blob(object_name, **options)
        blob.download_to_filename(filename)

        return filename


class MAS(Clouds):
    def __init__(self, **options):
        """
        @TODO: add documentation
        """
        try:
            from azure.storage.blob import BlockBlobService
        except ImportError as e:
            raise ImportError("No specified import name! make sure that you have installed the package via pip:\n\n"
                              "pip install azure-storage-blob")
        self.block_blob_service = BlockBlobService(**options)

    def upload_directory(self, directory, **options):
        container = options.pop('container', None)
        if container is None:
            raise ValueError('You should pass a container name')

        files = [f for f in listdir(directory) if isfile(join(directory, f))]

        try:
            for file in files:
                self.block_blob_service.create_blob_from_path(container, file, join(directory, file))
        except:
            error = "An error occurred while uploading the directory"
            logging.error(error)
            raise RuntimeError(error)

    def download(self, filename=None, **options):
        container = options.pop('container', None)
        blob = options.pop('blob', None)

        if container is None or blob is None:
            raise ValueError('You should pass a container name and a blob name')

        if filename is None:
            with tempfile.NamedTemporaryFile(prefix=basename(blob), delete=False) as tmp:
                filename = tmp.name

        try:
            self.block_blob_service.get_blob_to_path(container, blob, filename)
            logging.info("The " + filename + " file was downloaded")
        except:
            error = "An error occurred while downloading the file"
            logging.error(error)
            raise RuntimeError(error)

        return filename


class CloudManager:
    def __init__(self, filename: str = None):
        """
        @TODO: add documentation
        """
        self.filename = filename
        self.clouds = []

    def add(self, cloud: Clouds, **options):
        self.clouds.append((cloud, options))
        return self

    def transfer(self, method, path):
        for cloud in self.clouds:
            getattr(cloud[0], method)(path, **cloud[1])


__all__ = [
    'Clouds',
    'CloudManager',
    'S3',
    'GCS',
    'MAS'
]



"""
ffmpeg_streaming.media
~~~~~~~~~~~~

Auto Generate Representation and Representation Object


:copyright: (c) 2020 by Amin Yazdanpanah.
:website: https://www.aminyazdanpanah.com
:email: contact@aminyazdanpanah.com
:license: MIT, see LICENSE for more details.
"""
from ._format import Format
from ._media_property import Size, Bitrate

MINIMUM_BITRATE = 65536


class Representation:
    def __init__(self, size: Size, bitrate: Bitrate, **options):
        """
        @TODO: add documentation
        """
        self.size = size
        self.bitrate = bitrate
        self.options = options


def min_bitrate(bitrate: int) -> int:
    """
    @TODO: add documentation
    """
    if bitrate < MINIMUM_BITRATE:
        return MINIMUM_BITRATE

    return bitrate


def reduce_bitrate(bitrate: Bitrate, divide: int) -> Bitrate:
    """
    @TODO: add documentation
    """
    if divide == 1:
        return bitrate

    divide = 1 + divide / 2

    overall = min_bitrate(int(bitrate.overall_ / divide))
    video = min_bitrate(int(bitrate.video_ / divide)) if bitrate.video_ is not None and bitrate.video_ != 0 else None
    audio = min_bitrate(int(bitrate.audio_ / divide)) if bitrate.audio_ is not None and bitrate.audio_ != 0 else None

    return Bitrate(video, audio, overall)


def cal_bitrate(bitrate, org_bitrate: Bitrate, index: int) -> Bitrate:
    return bitrate[index - 1] if bitrate is not None else reduce_bitrate(org_bitrate, index)


class AutoRep(object):
    def __init__(self, original_size: Size, original_bitrate: Bitrate, _format: Format,
                 heights: list = None, bitrate: list = None, include_original: bool = True):
        """
        @TODO: add documentation
        """
        self.include_original = include_original
        self.original_bitrate = original_bitrate
        self.original_size = original_size
        self._format = _format
        self.heights = heights if heights is not None else [2160, 1440, 1080, 720, 480, 360, 240, 144]
        self.bitrate = bitrate
        self.is_default = True if heights is not None and bitrate is not None else False

        if heights is not None and bitrate is not None and len(heights) != len(bitrate):
            raise ValueError("The length of heights list must the same as length of bitrate")

    def __iter__(self):
        """
        @TODO: add documentation
        """
        if not self.is_default:
            height = self.original_size.ratio.calculate_height(self.original_size.width, self._format.multiply())
            original = [height] if self.include_original else []
            self.heights = original + list(filter(lambda x: x < height, self.heights))

        self.index = 0

        return self

    def __next__(self):
        """
        @TODO: add documentation
        """
        if self.index < len(self.heights):
            height = self.heights[self.index]
            width = self.original_size.ratio.calculate_width(height, self._format.multiply())
            self.index += 1

            return Representation(Size(width, height), cal_bitrate(self.bitrate, self.original_bitrate, self.index))
        else:
            raise StopIteration


__all__ = [
    'Representation'
]



"""
ffmpeg_streaming.process
~~~~~~~~~~~~

Run FFmpeg commands and monitor FFmpeg


:copyright: (c) 2020 by Amin Yazdanpanah.
:website: https://www.aminyazdanpanah.com
:email: contact@aminyazdanpanah.com
:license: MIT, see LICENSE for more details.
"""

import shlex
import subprocess
import threading
import logging
import time

from ffmpeg_streaming._hls_helper import HLSKeyInfoFile
from ffmpeg_streaming._utiles import get_time, time_left


def _p_open(commands, **options):
    """
    @TODO: add documentation
    """
    logging.info("ffmpeg running command: {}".format(commands))
    return subprocess.Popen(shlex.split(commands), **options)


class Process(object):
    out = None
    err = None

    def __init__(self, media, commands: str, monitor: callable = None, **options):
        """
        @TODO: add documentation
        """
        self.is_monitor = False
        self.input = options.pop('input', None)
        self.timeout = options.pop('timeout', None)
        default_proc_opts = {
            'stdin': None,
            'stdout': subprocess.PIPE,
            'stderr': subprocess.STDOUT,
            'universal_newlines': False
        }
        default_proc_opts.update(options)
        options.update(default_proc_opts)
        if callable(monitor) or isinstance(getattr(media, 'key_rotation'), HLSKeyInfoFile):
            self.is_monitor = True
            options.update({
                'stdin': subprocess.PIPE,
                'universal_newlines': True
            })

        self.process = _p_open(commands, **options)
        self.media = media
        self.monitor = monitor

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.process.kill()

    def _monitor(self):
        """
        @TODO: add documentation
        """
        duration = 1
        _time = 0
        log = []
        start_time = time.time()

        while True:
            line = self.process.stdout.readline().strip()
            if line == '' and self.process.poll() is not None:
                break

            if line != '':
                log += [line]

            if isinstance(getattr(self.media, 'key_rotation'), HLSKeyInfoFile):
                getattr(self.media, 'key_rotation').rotate_key(line)

            if callable(self.monitor):
                duration = get_time('Duration: ', line, duration)
                _time = get_time('time=', line, _time)
                self.monitor(line, duration, _time, time_left(start_time, _time, duration), self.process)

        Process.out = log

    def _thread_mon(self):
        """
        @TODO: add documentation
        """
        thread = threading.Thread(target=self._monitor)
        thread.start()

        thread.join(self.timeout)
        if thread.is_alive():
            self.process.terminate()
            thread.join()
            error = 'Timeout! exceeded the timeout of {} seconds.'.format(str(self.timeout))
            logging.error(error)
            raise RuntimeError(error)

    def run(self):
        """
        @TODO: add documentation
        """
        if self.is_monitor:
            self._thread_mon()
        else:
            Process.out, Process.err = self.process.communicate(self.input, self.timeout)

        if self.process.poll():
            error = str(Process.err) if Process.err else str(Process.out)
            logging.error('ffmpeg failed to execute command: {}'.format(error))
            raise RuntimeError('ffmpeg failed to execute command: ', error)

        logging.info("ffmpeg executed command successfully")

        return Process.out, Process.err



"""
ffmpeg_streaming.media
~~~~~~~~~~~~

Video and audio formats


:copyright: (c) 2020 by Amin Yazdanpanah.
:website: https://www.aminyazdanpanah.com
:email: contact@aminyazdanpanah.com
:license: MIT, see LICENSE for more details.
"""
import abc


MULTIPLY_BY_ONE = 1
MULTIPLY_BY_TWO = 2
MULTIPLY_BY_FOUR = 4
MULTIPLY_BY_Eight = 8
MULTIPLY_BY_SIXTEEN = 16
MULTIPLY_BY_THIRTY_TWO = 32


def _verify_codecs(codec, codecs):
    if codec is None:
        return
    elif codec not in codecs:
        ValueError("The codec is not available!")
    else:
        return str(codec)


class Format(abc.ABC):
    """
    @TODO: add documentation
    """
    def __init__(self, video: str, audio: str, **codec_options):
        self.video = video
        self.audio = audio
        self.codec_options = codec_options

    @property
    def all(self) -> dict:
        args = {
            'c:v': self.video,
            'c:a': self.audio,
        }
        args.update(self.get_codec_options())

        return args

    @abc.abstractmethod
    def multiply(self) -> int:
        pass

    @abc.abstractmethod
    def get_codec_options(self) -> dict:
        pass


class H264(Format):
    def __init__(self, video: str = "libx264", audio: str = 'aac', **codec_options):
        """
        @TODO: add documentation
        """
        videos = ['libx264', 'h264', 'h264_afm', 'h264_nvenc']
        audios = ['copy', 'aac', 'libvo_aacenc', 'libfaac', 'libmp3lame', 'libfdk_aac']

        super(H264, self).__init__(_verify_codecs(video, videos), _verify_codecs(audio, audios), **codec_options)

    def multiply(self) -> int:
        return MULTIPLY_BY_TWO

    def get_codec_options(self) -> dict:
        """
        set the default value of h264 codec options and update the value with the specified value by user
        see https://ffmpeg.org/ffmpeg-codecs.html#Options-28 for more information about options
        :return: dict
        """
        h264_codec_options = {
            'bf':           1,
            'keyint_min':   25,
            'g':            250,
            'sc_threshold': 40
        }

        h264_codec_options.update(self.codec_options)

        return h264_codec_options


class HEVC(Format):
    """
    @TODO: add documentation
    """
    def __init__(self, video: str = "libx265", audio: str = 'aac', **codec_options):
        videos = ['libx265', 'h265']
        audios = ['copy', 'aac', 'libvo_aacenc', 'libfaac', 'libmp3lame', 'libfdk_aac']

        super(HEVC, self).__init__(_verify_codecs(video, videos), _verify_codecs(audio, audios), **codec_options)

    def multiply(self) -> int:
        return MULTIPLY_BY_TWO

    def get_codec_options(self) -> dict:
        """
        set the default value of hevc(h265) codec options and update the value with the specified value by user
        see https://ffmpeg.org/ffmpeg-codecs.html#Options-29 for more information about options
        :return: dict
        """
        h265_codec_options = {
            'keyint_min':   25,
            'g':            250,
            'sc_threshold': 40
        }

        h265_codec_options.update(self.codec_options)

        return h265_codec_options


class VP9(Format):
    """
    @TODO: add documentation
    """
    def __init__(self, video: str = "libvpx-vp9", audio: str = 'aac', **codec_options):
        videos = ['libvpx', 'libvpx-vp9']
        audios = ['copy', 'aac', 'libvo_aacenc', 'libfaac', 'libmp3lame', 'libfdk_aac']

        super(VP9, self).__init__(_verify_codecs(video, videos), _verify_codecs(audio, audios), **codec_options)

    def multiply(self) -> int:
        return MULTIPLY_BY_TWO

    def get_codec_options(self) -> dict:
        """
        set the default value of vp9 codec options and update the value with the specified value by user
        see https://ffmpeg.org/ffmpeg-codecs.html#Options-26 for more information about options
        :return: dict
        """
        vp9_codec_options = {}

        vp9_codec_options.update(self.codec_options)

        return vp9_codec_options


class Formats:
    @staticmethod
    def h264(video: str = "libx264", audio: str = 'aac', **codec_options) -> Format:
        """
        @TODO: add documentation
        """
        return H264(video, audio, **codec_options)

    @staticmethod
    def hevc(video: str = "libx265", audio: str = 'aac', **codec_options) -> Format:
        """
        @TODO: add documentation
        """
        return HEVC(video, audio, **codec_options)

    @staticmethod
    def vp9(video: str = "libvpx-vp9", audio: str = 'aac', **codec_options) -> Format:
        """
        @TODO: add documentation
        """
        return VP9(video, audio, **codec_options)


__all__ = [
    'Format',
    'Formats'
]



"""
ffmpeg_streaming._ffprobe
~~~~~~~~~~~~

Probe the video


:copyright: (c) 2020 by Amin Yazdanpanah.
:website: https://www.aminyazdanpanah.com
:email: contact@aminyazdanpanah.com
:license: MIT, see LICENSE for more details.
"""

import json
import logging
import subprocess

from ._media_property import Size, Bitrate


class Streams:
    def __init__(self, streams):
        self.streams = streams

    def video(self, ignore_error=True):
        """
        @TODO: add documentation
        """
        return self._get_stream('video', ignore_error)

    def audio(self, ignore_error=True):
        """
        @TODO: add documentation
        """
        return self._get_stream('audio', ignore_error)

    def first_stream(self):
        """
        @TODO: add documentation
        """
        return self.streams[0]

    def all(self):
        """
        @TODO: add documentation
        """
        return self.streams

    def videos(self):
        """
        @TODO: add documentation
        """
        return self._get_streams('video')

    def audios(self):
        """
        @TODO: add documentation
        """
        return self._get_streams('audio')

    def _get_stream(self, media, ignore_error):
        """
        @TODO: add documentation
        """
        media_attr = next((stream for stream in self.streams if stream['codec_type'] == media), None)
        if media_attr is None and not ignore_error:
            raise ValueError('No {} stream found'.format(str(media)))
        return media_attr if media_attr is not None else {}

    def _get_streams(self, media):
        """
        @TODO: add documentation
        """
        for stream in self.streams:
            if stream['codec_type'] == media:
                yield stream


class FFProbe:
    def __init__(self, filename, cmd='ffprobe'):
        """
        @TODO: add documentation
        """
        commands = [cmd, '-show_format', '-show_streams', '-of', 'json', filename]
        logging.info("ffprobe running command: {}".format(" ".join(commands)))
        process = subprocess.Popen(commands, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        self.out, err = process.communicate()
        if process.returncode != 0:
            logging.error(str(self.out) + str(err))
            raise RuntimeError('ffprobe', self.out, err)
        logging.info("ffprobe executed command successfully!")

    def streams(self):
        """
        @TODO: add documentation
        """
        return Streams(json.loads(self.out.decode('utf-8'))['streams'])

    def format(self):
        """
        @TODO: add documentation
        """
        return json.loads(self.out.decode('utf-8'))['format']

    def all(self):
        """
        @TODO: add documentation
        """
        return json.loads(self.out.decode('utf-8'))

    def save_as_json(self, path):
        """
        @TODO: add documentation
        """
        with open(path, 'w') as probe:
            probe.write(self.out.decode('utf-8'))

    @property
    def video_size(self) -> Size:
        """
        @TODO: add documentation
        """
        width = int(self.streams().video().get('width', 0))
        height = int(self.streams().video().get('height', 0))

        if width == 0 or height == 0:
            raise RuntimeError('It could not determine the value of width/height')

        return Size(width, height)

    @property
    def bitrate(self, _type: str = "k") -> Bitrate:
        """
        @TODO: add documentation
        """
        overall = int(self.format().get('bit_rate', 0))
        video = int(self.streams().video().get('bit_rate', 0))
        audio = int(self.streams().audio().get('bit_rate', 0))

        if overall == 0:
            raise RuntimeError('It could not determine the value of bitrate')

        return Bitrate(video, audio, overall, type=_type)


__all__ = [
    'FFProbe'
]



"""
ffmpeg_streaming.media
~~~~~~~~~~~~

HLS helper


:copyright: (c) 2020 by Amin Yazdanpanah.
:website: https://www.aminyazdanpanah.com
:email: contact@aminyazdanpanah.com
:license: MIT, see LICENSE for more details.
"""
import os
import uuid
from secrets import token_bytes, token_hex

from ffmpeg_streaming._utiles import mkdir, get_path_info


class HLSKeyInfoFile:
    def __init__(self, key_info_file_path: str, path: str, url: str, period: int = 0, needle: str = '', length: int = 16):
        """
        @TODO: add documentation
        """
        self.needle = needle
        self.period = period
        self.segments = []
        self.length = length
        self.url = self.c_url = url
        self.path = self.c_path = path
        mkdir(os.path.dirname(path))
        self.key_info_file_path = key_info_file_path

    def __str__(self):
        """
        @TODO: add documentation
        """
        self.generate()
        return self.key_info_file_path

    def generate(self):
        """
        @TODO: add documentation
        """
        self.generate_key()
        self.update_key_info_file()

    def generate_key(self):
        """
        @TODO: add documentation
        """
        with open(self.path, 'wb') as key:
            key.write(token_bytes(self.length))

    def update_key_info_file(self):
        """
        @TODO: add documentation
        """
        with open(self.key_info_file_path, 'w') as key_info_file:
            key_info_file.write("\n".join([self.url, self.path, token_hex(self.length)]))

    def update_suffix(self):
        """
        @TODO: add documentation
        """
        unique = uuid.uuid4()
        self.path = self.c_path + "-" + str(unique)
        self.url = self.c_url + "-" + str(unique)

    def rotate_key(self, line: str):
        """
        @TODO: add documentation
        """
        if self.needle in line and line not in self.segments:
            self.segments.append(line)
            if len(self.segments) % self.period == 0:
                self.update_suffix()
                self.generate()

def sub_info(rep, sub_path) -> list:
    """
    Returns the subtitle information to be added to manifest.

    Parameters
    ----------
    rep : Representation
    sub_path : subtitle manifest file name
    """
    tag = '#EXT-X-MEDIA:'
    info = [
        f'TYPE=SUBTITLES',
        f'GROUP-ID="subs"',
        f'NAME="subtitles"',
        f'URI="'+sub_path+'"'
    ]
    return [tag + ",".join(info)]

def stream_info(rep, sub_exists) -> list:
    """
    @TODO: add documentation
    """
    tag = '#EXT-X-STREAM-INF:'
    info = [
        f'BANDWIDTH={rep.bitrate.calc_overall}',
        f'RESOLUTION={rep.size}',
        f'NAME="{rep.size.height}"'
    ]
    if sub_exists:
        info.append(f'SUBTITLES="subs"')
    custom = rep.options.pop('stream_info', [])

    return [tag + ",".join(info + custom)]


class HLSMasterPlaylist:
    def __init__(self, media):
        """
        @TODO: add documentation
        """
        self.media = media

    @classmethod
    def generate(cls, media, path=None):
        if path is None:
            path = "{}.m3u8".format(os.path.join(*get_path_info(media.output_)))
        with open(path, 'w', encoding='utf-8') as playlist:
            playlist.write(cls(media)._content())

    def _content(self) -> str:
        """
        @TODO: add documentation
        """
        content = ['#EXTM3U'] + self._get_version() + self.media.options.get('description', [])

        for rep in self.media.reps:
            sub_exists=os.path.isfile(os.path.dirname(self.media.output_)+'/'+self.sub_path(rep)[0])
            if(sub_exists):
                content += sub_info(rep,self.sub_path(rep)[0]) + stream_info(rep,sub_exists) + self.stream_path(rep)
            else:
                content += stream_info(rep,sub_exists) + self.stream_path(rep)

        return "\n".join(content)

    def _get_version(self) -> list:
        """
        @TODO: add documentation
        """
        version = "7" if self.media.options.get('hls_segment_type', '') == 'fmp4' else "3"
        return ['#EXT-X-VERSION:' + version]

    def stream_path(self, rep):
        """
        @TODO: add documentation
        """
        return ["{}_{}p.m3u8".format(os.path.basename(self.media.output_).split('.')[0], rep.size.height)]

    def sub_path(self, rep):
        """
        Returns the subtitles maifest file name.

        Parameters
        ----------
        rep : Representation
        """
        return ["{}_{}p_vtt.m3u8".format(os.path.basename(self.media.output_).split('.')[0], rep.size.height)]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="python-ffmpeg-video-streaming",
    version="0.1.15",
    author="Amin Yazdanpanah",
    author_email="contact@aminyazdanpanah.com",
    description="Package media content for online streaming(DASH and HLS) using ffmpeg",
    long_description=long_description,
    project_urls={
        "Bug Tracker": "https://github.com/aminyazdanpanah/python-ffmpeg-video-streaming/issues",
        "Documentation": "https://video.aminyazdanpanah.com/python",
        "Source Code": "https://github.com/aminyazdanpanah/python-ffmpeg-video-streaming",
    },
    long_description_content_type="text/markdown",
    url="https://github.com/aminyazdanpanah/python-ffmpeg-video-streaming",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)


"""
ffmpeg_streaming.streams
~~~~~~~~~~~~

Useful methods


:copyright: (c) 2020 by Amin Yazdanpanah.
:website: https://www.aminyazdanpanah.com
:email: contact@aminyazdanpanah.com
:license: MIT, see LICENSE for more details.
"""
import logging
import os
import re
import time
import warnings
from sys import platform


def get_path_info(path):
    """
    @TODO: add documentation
    """
    dirname = os.path.dirname(path)
    name = str(os.path.basename(path).rsplit('.', 1)[0])

    return dirname, name


def mkdir(dirname: str) -> None:
    """
    @TODO: add documentation
    """
    try:
        os.makedirs(dirname)
    except OSError as exc:
        logging.info(exc)
        pass


def rm(path: str) -> None:
    """
    @TODO: add documentation
    """
    try:
        os.remove(path)
    except OSError as exc:
        logging.info(exc)
        pass


def clean_args(args: list) -> list:
    """
    @TODO: add documentation
    """
    clean_args_ = []
    for arg in args:
        if " " in arg:
            arg = '"' + arg + '"'
        clean_args_.append(arg.replace("\\", "/").replace("__COLON__", ":"))

    return clean_args_


def convert_to_sec(time):
    """
    @TODO: add documentation
    """
    h, m, s = time.split(":")
    return int(h) * 3600 + int(m) * 60 + int(s)


def get_time(key, string, default):
    """
    @TODO: add documentation
    """
    time = re.search('(?<={})\w+:\w+:\w+'.format(key), string)
    return convert_to_sec(time.group(0)) if time else default


def time_left(start_time, unit, total):
    """
    @TODO: add documentation
    """
    if unit != 0:
        diff_time = time.time() - start_time
        return total * diff_time / unit - diff_time
    else:
        return 0


def deprecated(func):
    """
    @TODO: add documentation
    """
    def deprecated_fun(*args, **kwargs):
        warnings.warn('The {} method is deprecated and will be removed in a future release'.format(func.__name__)
                      , DeprecationWarning, stacklevel=2)
        return func(*args, **kwargs)
    return deprecated_fun


def get_os():
    """
    @TODO: add documentation
    """
    if platform == "linux" or platform == "linux2":
        os_name = 'linux'
    elif platform == "darwin":
        os_name = 'os_x'
    elif platform == "win32" or platform == "Windows":
        os_name = 'windows'
    else:
        os_name = 'unknown'

    return os_name


def cnv_options_to_args(options: dict):
    """
    @TODO: add documentation
    """
    args = []
    for k, v in options.items():
        args.append('-{}'.format(k))
        if v is not None:
            args.append('{}'.format(v))

    return args



"""
ffmpeg_streaming.build_args
~~~~~~~~~~~~

Build a command for the FFmpeg


:copyright: (c) 2020 by Amin Yazdanpanah.
:website: https://www.aminyazdanpanah.com
:email: contact@aminyazdanpanah.com
:license: MIT, see LICENSE for more details.
"""
import sys

from ._utiles import cnv_options_to_args, get_path_info, clean_args

# DASH default values
USE_TIMELINE = 1
USE_TEMPLATE = 1

# HLS default values
HLS_LIST_SIZE = 0
HLS_TIME = 10
HLS_ALLOW_CACHE = 1


def _stream2file(stream2file):
    """
    @TODO: add documentation
    """
    args = stream2file.format.all
    args.update({'c': 'copy'})
    args.update(stream2file.options)

    return cnv_options_to_args(args) + [stream2file.output_]


def _get_audio_bitrate(rep, index: int = None):
    """
    @TODO: add documentation
    """
    if rep.bitrate.audio_ is not None and rep.bitrate.audio_ != 0:
        opt = 'b:a' if index is None else 'b:a:' + str(index)
        return {opt: rep.bitrate.audio}

    return {}


def _get_dash_stream(key, rep):
    """
    @TODO: add documentation
    """
    args = {
        'map': 0,
        's:v:' + str(key): rep.size,
        'b:v:' + str(key): rep.bitrate.calc_video()
    }
    args.update(_get_audio_bitrate(rep, key))
    args.update(rep.options)
    return cnv_options_to_args(args)


def _dash(dash):
    """
    @TODO: add documentation
    """
    dirname, name = get_path_info(dash.output_)
    _args = dash.format.all
    _args.update({
        'use_timeline':     USE_TIMELINE,
        'use_template':     USE_TEMPLATE,
        'init_seg_name':    '{}_init_$RepresentationID$.$ext$'.format(name),
        "media_seg_name":   '{}_chunk_$RepresentationID$_$Number%05d$.$ext$'.format(name),
        'f': 'dash'
    })
    _args.update(dash.options)
    args = cnv_options_to_args(_args)

    for key, rep in enumerate(dash.reps):
        args += _get_dash_stream(key, rep)

    return args + ['-strict', '-2', '{}/{}.mpd'.format(dirname, name)]


def _hls_seg_ext(hls):
    """
    @TODO: add documentation
    """
    return 'm4s' if hls.options.get('hls_segment_type', '') == 'fmp4' else 'ts'


def _get_hls_stream(hls, rep, dirname, name):
    """
    @TODO: add documentation
    """
    args = hls.format.all
    args.update({
        'hls_list_size':            HLS_LIST_SIZE,
        'hls_time':                 HLS_TIME,
        'hls_allow_cache':          HLS_ALLOW_CACHE,
        'hls_segment_filename':     "{}/{}_{}p_%04d.{}".format(dirname, name, rep.size.height, _hls_seg_ext(hls)),
        'hls_fmp4_init_filename':   "{}_{}p_init.mp4".format(name, rep.size.height),
        's:v':                      rep.size,
        'b:v':                      rep.bitrate.calc_video()
    })
    args.update(_get_audio_bitrate(rep))
    args.update(rep.options)
    args.update({'strict': '-2'})
    args.update(hls.options)

    return cnv_options_to_args(args) + ["{}/{}_{}p.m3u8".format(dirname, name, str(rep.size.height))]


def _hls(hls):
    """
    @TODO: add documentation
    """
    dirname, name = get_path_info(hls.output_)
    streams = []
    for key, rep in enumerate(hls.reps):
        if key > 0:
            streams += input_args(hls)
        streams += _get_hls_stream(hls, rep, dirname, name)

    return streams


def stream_args(media):
    """
    @TODO: add documentation
    """
    return getattr(sys.modules[__name__], "_%s" % type(media).__name__.lower())(media)


def input_args(media):
    inputs = []
    for key, _input in enumerate(media.media.inputs.inputs):
        _input = dict(_input)
        _input.pop('is_tmp', None)
        if key > 0:
            _input.pop('y', None)
        inputs += cnv_options_to_args(_input)

    return inputs


def command_builder(ffmpeg_bin: str, media):
    """
    @TODO: add documentation
    """
    return " ".join(clean_args([ffmpeg_bin] + input_args(media) + stream_args(media)))
